package server

import (
	"xbus.io/xbus/services"
	"xbus.io/xbus/services/broker"
	"xbus.io/xbus/services/clientapi"
	"xbus.io/xbus/services/control"
	"xbus.io/xbus/services/core"
	"xbus.io/xbus/services/director"
	"xbus.io/xbus/services/envelopestorage"
	"xbus.io/xbus/services/metrics"
	"xbus.io/xbus/services/postmortem"
	"xbus.io/xbus/services/processstate"
	"xbus.io/xbus/services/ravencapture"
	"xbus.io/xbus/services/registration"
	"xbus.io/xbus/services/sentinel"
)

func (s *Server) servicesOptions() services.Options {
	return services.Options{
		Conn:   s.natsConn,
		Logger: s.log,
	}
}

func (s *Server) initServiceRavenCapture() *ravencapture.Service {
	return ravencapture.Options{
		Options:     s.servicesOptions(),
		RavenClient: s.ravenClient,
	}.NewService()
}

func (s *Server) initServiceBroker() *broker.Service {
	return broker.New(broker.Options{
		Options:         s.servicesOptions(),
		ActorStorage:    s.actorStorage,
		EnvelopeStorage: s.envelopeStorage,
		LogStorage:      s.logStorage,
		ProcessManager:  s.processManager,
	})
}

func (s *Server) initServiceClientAPI() *clientapi.Service {
	return clientapi.NewService(clientapi.Options{
		Options:        s.servicesOptions(),
		AccountStorage: s.accountStorage,
		ActorStorage:   s.actorStorage,
		SessionStorage: s.sessionStorage,
	})
}

func (s *Server) initServiceControl() *control.Service {
	return control.NewService(
		control.Options{
			Options:         s.servicesOptions(),
			ServerCert:      s.clientCA,
			ServerKey:       s.clientCAKey,
			PipelineManager: s.pipelineManager,
			ProcessStorage:  s.processStorage,
			ProcessManager:  s.processManager,
			AccountStorage:  s.accountStorage,
			PasswordStorage: s.passwordStorage,
			ActorStorage:    s.actorStorage,
			EnvelopeStorage: s.envelopeStorage,
			LogStorage:      s.logStorage,
			JobStorage:      s.jobStorage,
		})
}

func (s *Server) initServiceCore() *core.Service {
	return core.Options{
		Options:         s.servicesOptions(),
		ActorStorage:    s.actorStorage,
		ProcessStorage:  s.processStorage,
		ProcessManager:  s.processManager,
		PipelineManager: s.pipelineManager,
		LogStorage:      s.logStorage,
		EnvelopeStorage: s.envelopeStorage,
	}.NewService()
}

func (s *Server) initServiceDirector() *director.Service {
	return director.New(director.Options{
		Options:         s.servicesOptions(),
		ActorStorage:    s.actorStorage,
		EnvelopeStorage: s.envelopeStorage,
		LogStorage:      s.logStorage,
		ProcessManager:  s.processManager,
		JobStorage:      s.jobStorage,
	})
}

func (s *Server) initServiceEnvelopeStorage() *envelopestorage.Service {
	opts := envelopestorage.Options{
		Options:         s.servicesOptions(),
		EnvelopeStorage: s.envelopeStorage,
		JobStorage:      s.jobStorage,
	}
	return opts.NewService()
}

func (s *Server) initServiceMetrics() *metrics.Service {
	opts := metrics.Options{
		Options: s.servicesOptions(),

		ServerCA: s.serverCert.Raw,
		ClientCA: s.clientCA.Raw,

		RavenClient: s.ravenClient,
	}
	return opts.NewService()
}

func (s *Server) initServiceProcessState() *processstate.Service {
	return processstate.Options{
		Options:         s.servicesOptions(),
		LogStorage:      s.logStorage,
		ProcessManager:  s.processManager,
		EnvelopeStorage: s.envelopeStorage,
		ActorStorage:    s.actorStorage,
	}.NewService()
}

func (s *Server) initServicePostmortem() *postmortem.Service {
	return postmortem.Options{
		Options:        s.servicesOptions(),
		LogStorage:     s.logStorage,
		ProcessManager: s.processManager,
	}.NewService()
}

func (s *Server) initServiceRegistration() *registration.Service {
	return registration.NewService(
		registration.Options{
			Options:        s.servicesOptions(),
			AccountStorage: s.accountStorage,
			ActorStorage:   s.actorStorage,
			ServerCA:       s.serverCert.Raw,
		})
}

func (s *Server) initServiceSentinel() *sentinel.Service {
	return sentinel.Options{
		Options:      s.servicesOptions(),
		ActorStorage: s.actorStorage,
	}.NewService()
}

func (s *Server) setupServices() {
	s.ServiceManager = services.NewManager(s.log)

	if s.ravenClient != nil {
		s.ServiceManager.Add(s.initServiceRavenCapture())
	}

	s.ServiceManager.Add(s.initServiceCore())

	if s.options.ServeSentinel {
		s.ServiceManager.Add(s.initServiceSentinel())
	}

	if s.options.ServePostmortem {
		s.ServiceManager.Add(s.initServicePostmortem())
	}

	if s.options.ServeBroker {
		s.ServiceManager.Add(s.initServiceBroker())
	}

	if s.options.ServeClientAPI {
		s.ServiceManager.Add(s.initServiceClientAPI())
	}

	if s.options.ServeControl {
		s.ServiceManager.Add(s.initServiceControl())
	}

	if s.options.ServeDirector {
		s.ServiceManager.Add(s.initServiceDirector())
	}

	if s.options.ServeEnvelopeStorage {
		s.ServiceManager.Add(s.initServiceEnvelopeStorage())
	}

	if s.options.ServeProcessState {
		s.ServiceManager.Add(s.initServiceProcessState())
	}

	if s.options.ServeRegistration {
		s.ServiceManager.Add(s.initServiceRegistration())
	}

	if s.options.MetricsHTTPPort != 0 && s.options.ServeMetrics {
		s.ServiceManager.Add(s.initServiceMetrics())
	}
}
