// Copyright 2019 Orus.io. All rights reserved.

package metrics

import (
	"context"
	"crypto/x509"
	"errors"
	"fmt"
	"math"
	"sync"
	"time"

	"github.com/getsentry/raven-go"
	"github.com/prometheus/client_golang/prometheus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/api/control"
	xbus_tools "xbus.io/go-xbus/tools"
	"xbus.io/xbus/services"
	"xbus.io/xbus/services/tools"
)

func min(a, b uint) uint {
	if a < b {
		return a
	}
	return b
}

const (
	// MetricsQueue is the name we use to queueSub to a queue with the metrics
	MetricsQueue = "metrics_queue"
)

// Options is used to create a Service
type Options struct {
	services.Options

	CertificateExpirationCheckInterval    time.Duration
	CertificateExpirationNotifyInterval   time.Duration
	CertificateExpirationWarningThreshold time.Duration
	CertificateExpirationErrorThreshold   time.Duration

	ServerCA    []byte
	ClientCA    []byte
	RavenClient *raven.Client
}

// NewService returns a new Service
func (o Options) NewService() *Service {
	if o.CertificateExpirationCheckInterval == 0 {
		o.CertificateExpirationCheckInterval = time.Minute
	}
	if o.CertificateExpirationNotifyInterval == 0 {
		o.CertificateExpirationNotifyInterval = time.Hour
	}
	if o.CertificateExpirationWarningThreshold == 0 {
		o.CertificateExpirationWarningThreshold = time.Hour * 24 * 28
	}
	if o.CertificateExpirationErrorThreshold == 0 {
		o.CertificateExpirationErrorThreshold = time.Hour * 24 * 7
	}
	serverCA, err := x509.ParseCertificate(o.ServerCA)
	if err != nil {
		o.Logger.Errorf("Could not parse server ca: %s", err)
		return nil
	}
	clientCA, err := x509.ParseCertificate(o.ClientCA)
	if err != nil {
		o.Logger.Errorf("Could not parse client ca: %s", err)
		return nil
	}
	return &Service{
		BaseService: services.NewBaseService("metrics", o.Options),
		Opts:        o,

		serverCAValidNotAfter: serverCA.NotAfter,
		clientCAValidNotAfter: clientCA.NotAfter,
	}
}

// New returns a new Service
func New(options Options) *Service {
	return options.NewService()
}

// Service is the metrics service implementation
// The metrics service is responsible for keeping track of global metrics
// of the bus
type Service struct {
	services.BaseService
	Opts Options

	ctx      context.Context
	stop     func()
	wg       sync.WaitGroup
	api      *api.Client
	control  *control.Client
	actorMap *tools.ActorMap

	serverCAValidNotAfter time.Time
	clientCAValidNotAfter time.Time
}

// Startup ...
func (s *Service) Startup() error {
	s.ctx, s.stop = context.WithCancel(context.Background())
	s.api = api.NewClient(s.Conn)
	s.control = control.NewClient(s.Conn)
	if m, err := tools.NewActorMap(s.ctx, s.api, s.Log); err == nil {
		s.actorMap = m
	} else {
		return err
	}
	s.updateAllCertificateExpiration()
	s.wg.Add(1)
	go s.run()
	return nil
}

func (s *Service) run() {
	defer s.wg.Done()
	for _, actorState := range s.actorMap.List() {
		updateActorOnlineStatus(actorState)
	}
	certificateExpirationCheckTicker := time.NewTicker(s.Opts.CertificateExpirationCheckInterval)
	defer certificateExpirationCheckTicker.Stop()

	certificateNotifyTicker := time.NewTicker(s.Opts.CertificateExpirationNotifyInterval)
	defer certificateNotifyTicker.Stop()

	for {
		select {
		case actorState := <-s.actorMap.Events:
			updateActorOnlineStatus(actorState)
		case <-certificateExpirationCheckTicker.C:
			s.updateAllCertificateExpiration()
		case <-certificateNotifyTicker.C:
			s.checkForExpirationNotify()
		case <-s.ctx.Done():
			return
		}
	}
}

// Shutdown the service
func (s *Service) Shutdown() error {
	s.stop()
	s.wg.Wait()
	s.actorMap.Close()
	return nil
}

// metrics
var actorOnlineStatus = prometheus.NewGaugeVec(prometheus.GaugeOpts{
	Name: "xbus_server_actor_online_status",
	Help: "Consumer/worker online status. 0 if the actor is offline, 1 if the actor is online",
}, []string{"name"})

func updateActorOnlineStatus(actorState tools.ActorState) {
	if actorState.Kind != api.Actor_EMITTER {
		var value float64
		if actorState.Online {
			value = 1.0
		}
		actorOnlineStatus.WithLabelValues(actorState.Name).Set(value)
	}
}

var accountDaysToCertificateExpiration = prometheus.NewGaugeVec(prometheus.GaugeOpts{
	Name: "xbus_server_account_days_to_certificate_expiration",
	Help: "Days until actor certificate expiration",
}, []string{"name"})

var serverDaysToCertificateExpiration = prometheus.NewGaugeVec(prometheus.GaugeOpts{
	Name: "xbus_server_days_to_certificate_expiration",
	Help: "Days until a server certificate expiration",
}, []string{"name"})

func (s *Service) updateAllCertificateExpiration() {
	serverDaysToCertificateExpiration.
		WithLabelValues("server-ca").
		Set(time.Until(s.serverCAValidNotAfter).Hours() / 24)
	serverDaysToCertificateExpiration.
		WithLabelValues("client-ca").
		Set(time.Until(s.serverCAValidNotAfter).Hours() / 24)

	accountList, err := s.control.Account.List(control.AccountListRequest{})
	if err != nil {
		s.Opts.Logger.Errorf("Could not list accounts: %s", err)
		return
	}
	for _, account := range accountList.AccountList {
		s.updateAccountDaysToCertificateExpiration(account)
	}
}

// ErrNoCert is returned by accountCertificateValidUntil if the account has no
// certificate
var ErrNoCert = errors.New("account has not certificate")

func accountCertificateValidUntil(account api.Account) (time.Time, error) {
	if account.Cert == "" {
		return time.Time{}, ErrNoCert
	}
	cert, err := xbus_tools.ReadPEMCertificate(account.Cert)
	if err != nil {
		return time.Time{}, fmt.Errorf(
			"Could not parse account %s certificate: %s", account.Name, err)
	}
	return cert.NotAfter, nil
}

func (s *Service) updateAccountDaysToCertificateExpiration(account api.Account) {
	validUntil, err := accountCertificateValidUntil(account)
	if err != nil {
		if err != ErrNoCert {
			s.Opts.Logger.Errorf("%s", err)
		}
		return
	}
	days := time.Until(validUntil).Hours() / 24
	accountDaysToCertificateExpiration.WithLabelValues(account.Name).Set(days)
}

func (s *Service) notifyCertificate(name string, notAfter time.Time) {
	remains := time.Until(notAfter)
	remainDays := int(math.Floor(remains.Hours() / 24))
	level := raven.INFO

	if remains <= 0 {
		level = raven.FATAL
	} else if remains <= s.Opts.CertificateExpirationErrorThreshold {
		level = raven.ERROR
	} else if remains <= s.Opts.CertificateExpirationWarningThreshold {
		level = raven.WARNING
	}

	if level == raven.INFO {
		return
	}

	msg := fmt.Sprintf(
		"Certificate \"%s\" is about to expire. Time remaining: %d days", name, remainDays)

	if s.Opts.RavenClient != nil {
		paquet := raven.NewPacketWithExtra(msg, raven.Extra{"name": name, "days-remaining": remainDays})
		paquet.Level = level
		_, c := s.Opts.RavenClient.Capture(paquet, nil)
		if err := <-c; err != nil {
			s.Opts.Logger.Errorf(fmt.Sprintf("Error capturing sentry paquet: %s", err))
			s.Opts.Logger.Errorf(msg)
		}
	} else {
		if level == raven.WARNING {
			s.Opts.Logger.Noticef(msg)
		} else {
			s.Opts.Logger.Errorf(msg)
		}
	}
}

func (s *Service) checkForExpirationNotify() {
	s.notifyCertificate("server-ca", s.serverCAValidNotAfter)
	s.notifyCertificate("client-ca", s.clientCAValidNotAfter)

	accountList, err := s.control.Account.List(control.AccountListRequest{})
	if err != nil {
		s.Opts.Logger.Errorf("Could not list accounts: %s", err)
		return
	}
	for _, account := range accountList.AccountList {
		d, err := accountCertificateValidUntil(account)
		if err != nil {
			if err != ErrNoCert {
				s.Opts.Logger.Errorf("%s", err)
			}
			continue
		}
		s.notifyCertificate("Account "+account.Name, d)
	}
}

func init() {
	prometheus.MustRegister(accountDaysToCertificateExpiration)
	prometheus.MustRegister(actorOnlineStatus)
	prometheus.MustRegister(serverDaysToCertificateExpiration)
}
